import numpy as np
import torch
from torch import Tensor
from typing import Callable
from .utils import (
    hshap_features,
    shapley_matrix,
    mask_input_,
    mask_features_,
    mask_map_,
)


class Explainer:
    def __init__(
        self,
        model: Callable[[Tensor], Tensor],
        background: Tensor,
    ) -> None:
        self.model = model
        self.background = background
        self.size = (
            self.background.size(0),
            self.background.size(1),
            self.background.size(2),
        )
        self.gamma = 4
        self.features = hshap_features(self.gamma)
        self.W = shapley_matrix(self.gamma, device=background.device)

    def masked_input_(
        self,
        path: np.ndarray,
        root_input: Tensor,
        root_coords: np.ndarray,
    ) -> Tensor:
        mask_input_(
            input=root_input,
            path=path,
            background=self.background,
            root_coords=root_coords,
        )

        feature_mask = torch.zeros_like(root_input, dtype=torch.bool).repeat(
            len(self.features) + 1, 1, 1, 1
        )
        mask_features_(
            feature_mask=feature_mask,
            root_coords=root_coords,
        )
        m12 = torch.logical_or(feature_mask[1], feature_mask[2]).unsqueeze_(0)
        m13 = torch.logical_or(feature_mask[1], feature_mask[3]).unsqueeze_(0)
        m14 = torch.logical_or(feature_mask[1], feature_mask[4]).unsqueeze_(0)
        m23 = torch.logical_or(feature_mask[2], feature_mask[3]).unsqueeze_(0)
        m24 = torch.logical_or(feature_mask[2], feature_mask[4]).unsqueeze_(0)
        m34 = torch.logical_or(feature_mask[3], feature_mask[4]).unsqueeze_(0)
        m123 = torch.logical_or(m12, m13)
        m124 = torch.logical_or(m12, m14)
        m134 = torch.logical_or(m13, m14)
        m234 = torch.logical_or(m23, m24)
        m1234 = torch.logical_or(m123, m124)
        m = torch.cat(
            [
                feature_mask[:5],
                m12,
                m13,
                m14,
                m23,
                m24,
                m34,
                m123,
                m124,
                m134,
                m234,
                m1234,
            ]
        )
        return torch.where(m, root_input, self.background)

    def explain(
        self,
        x: Tensor,
        label: int,
        s: int,
        threshold_mode: str = "absolute",
        threshold: float = 0.0,
        softmax_activation: bool = True,
        batch_size: int = 2,
        binary_map: bool = False,
        roll_row: int = 0,
        roll_column: int = 0,
        **kwargs,
    ) -> Tensor:
        nodes = np.ones((1, 1, 4), dtype=np.bool_)
        scores = torch.ones((1,), device=self.background.device).float()
        root_coords = np.array(
            [[[0, 0], [self.size[1], self.size[2]]]], dtype=np.uint16
        )

        if roll_row != 0 or roll_column != 0:
            x = torch.roll(x, shifts=(-roll_row, -roll_column), dims=(1, 2))
        root_inputs = x.unsqueeze_(0)

        stop_l = int(np.log2(min(self.size[1], self.size[2]) / s)) + 2
        while nodes.shape[1] < stop_l:
            scores = scores.unsqueeze_(1).repeat((1, self.gamma))

            for batch_start_id in range(0, len(nodes), batch_size):
                batch = nodes[batch_start_id : batch_start_id + batch_size]
                batch_input = torch.empty_like(self.background).repeat(
                    len(batch), 2**self.gamma, 1, 1, 1
                )

                for n, node in enumerate(batch):
                    batch_input[n] = self.masked_input_(
                        node[-1],
                        root_inputs[batch_start_id + n],
                        root_coords[batch_start_id + n],
                    )

                    if roll_row != 0 or roll_column != 0:
                        batch_input[n] = torch.roll(
                            batch_input[n], shifts=(roll_row, roll_column), dims=(2, 3)
                        )

                F = self.model(
                    batch_input.view(
                        len(batch) * 2**self.gamma,
                        self.size[0],
                        self.size[1],
                        self.size[2],
                    ),
                    **kwargs,
                )

                if softmax_activation:
                    F = torch.nn.functional.softmax(F, dim=1)

                F = F[:, label]
                F = F.view((-1, 1, 2**self.gamma))
                scores[batch_start_id : batch_start_id + batch_size].mul_(
                    torch.matmul(F, self.W).squeeze_(1)
                )

            if threshold_mode == "absolute":
                masked_scores = scores > threshold
            if threshold_mode == "relative":
                t = torch.quantile(scores, threshold / 100, dim=None)
                if t <= 0:
                    masked_scores = scores > t
                else:
                    masked_scores = scores >= t

            i, j = masked_scores.nonzero(as_tuple=True)
            del masked_scores

            _i, _j = i.size(0), j.size(0)
            ic, jc = i.cpu(), j.cpu()

            if _i == 0 and _j == 0:
                raise ValueError("Could not find any important nodes.")
            if _i == 1 and _j == 1:
                nodes = np.concatenate(
                    (nodes[None, ic], self.features[None, jc]), axis=1, dtype=np.bool_
                )
                root_coords = root_coords[None, ic]
            else:
                nodes = np.concatenate(
                    (nodes[ic], self.features[jc]), axis=1, dtype=np.bool_
                )
                root_coords = root_coords[ic]

            scores = scores[i, j]
            root_inputs = root_inputs[i]

        saliency_map = torch.zeros(1, self.size[1], self.size[2])
        _scores = scores.tolist()

        for n, s, c in zip(nodes, _scores, root_coords):
            mask_map_(
                map=saliency_map,
                path=n[-1],
                score=s if not binary_map else 1,
                root_coords=c,
            )

        if roll_row != 0 or roll_column != 0:
            saliency_map = torch.roll(
                saliency_map, shifts=(roll_row, roll_column), dims=(1, 2)
            )
        return saliency_map

    def cycle_explain(
        self,
        x: Tensor,
        label: int,
        s: int,
        R: list[int],
        A: list[int],
        **kwargs,
    ):
        saliency_map = torch.zeros(1, self.size[1], self.size[2])
        for r in R:
            for a in A:
                if r == 0 and a != 0:
                    continue
                else:
                    roll_row = -int(r * np.sin(a))
                    roll_column = int(r * np.cos(a))

                    saliency_map += self.explain(
                        x.clone(),
                        label=label,
                        s=s,
                        roll_row=roll_row,
                        roll_column=roll_column,
                        **kwargs,
                    )
        saliency_map /= ((len(R) - 1) * len(A)) + 1
        saliency_map.squeeze_()
        return saliency_map



import torch
import numpy as np
from pytest import raises
from .src import Explainer
from .utils import hshap_features, shapley_matrix


def test_explainer():
    def model(x):
        t = x.flatten(start_dim=1).count_nonzero(dim=1)
        return torch.stack((t == 0, t > 0)).transpose(0, 1).type(torch.float)

    device = (
        torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
    )
    background = torch.zeros(3, 64, 64, device=device)

    x = torch.zeros(2, 3, 64, 64, device=device)
    x[1, :, 1, 1] = 1
    x[1, :, 2, 2] = 1
    assert torch.eq(model(x), torch.tensor([[1, 0], [0, 1]], device=device)).all()

    hexp = Explainer(model, background)
    assert (
        hexp.model == model
        and torch.eq(hexp.background, background).all()
        and hexp.size == (3, 64, 64)
        and hexp.gamma == 4
        and np.array_equal(hshap_features(hexp.gamma), hexp.features)
        and torch.eq(shapley_matrix(hexp.gamma, device), hexp.W).all()
    )

    with raises(ValueError, match="Could not find any important nodes."):
        hexp.explain(
            x=x[0],
            label=1,
            s=1,
            threshold_mode="absolute",
            threshold=0.0,
            softmax_activation=True,
            batch_size=2,
            binary_map=True,
            roll_row=0,
            roll_column=0,
        )

    expected_saliency_map = torch.zeros(1, 64, 64)
    expected_saliency_map[:, 1, 1] = 1
    expected_saliency_map[:, 2, 2] = 1
    assert torch.eq(
        expected_saliency_map,
        hexp.explain(
            x=x[1],
            label=1,
            s=1,
            threshold_mode="absolute",
            threshold=0.0,
            softmax_activation=True,
            batch_size=2,
            binary_map=True,
            roll_row=0,
            roll_column=0,
        ),
    ).all()

    expected_saliency_map = torch.zeros(1, 64, 64)
    expected_saliency_map[:, 1, 1] = 1
    expected_saliency_map[:, 2, 2] = 1
    assert torch.eq(
        expected_saliency_map,
        hexp.explain(
            x=x[1],
            label=1,
            s=1,
            threshold_mode="relative",
            threshold=70,
            softmax_activation=True,
            batch_size=2,
            binary_map=True,
            roll_row=0,
            roll_column=0,
        ),
    ).all()

    expected_saliency_map = torch.zeros(1, 64, 64)
    expected_saliency_map[:, 1:3, 1:3] = 1
    print(
        hexp.explain(
            x=x[1],
            label=1,
            s=2,
            threshold_mode="absolute",
            threshold=0.0,
            softmax_activation=False,
            batch_size=2,
            binary_map=True,
            roll_row=1,
            roll_column=1,
        )
    )
    assert torch.eq(
        expected_saliency_map,
        hexp.explain(
            x=x[1],
            label=1,
            s=2,
            threshold_mode="absolute",
            threshold=0.0,
            softmax_activation=False,
            batch_size=2,
            binary_map=True,
            roll_row=1,
            roll_column=1,
        ),
    ).all()

    expected_saliency_map = torch.zeros(1, 64, 64)
    expected_saliency_map[:, 1, 1] = 1
    expected_saliency_map[:, 2, 2] = 1
    assert torch.eq(
        expected_saliency_map,
        hexp.cycle_explain(
            x=x[1],
            label=1,
            s=1,
            R=[0, 1],
            A=[0, np.pi / 2, np.pi, 3 * np.pi / 2],
            threshold_mode="absolute",
            threshold=0.0,
            softmax_activation=False,
            batch_size=2,
            binary_map=True,
        ),
    ).all()



import numpy as np
import torch
from pytest import approx, raises
from .utils import (
    hshap_features,
    w,
    shapley_matrix,
    mask_features_,
    mask_input_,
    mask_map_,
)

factorial = np.math.factorial


def test_features():
    expected = np.array(
        [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]],
        dtype=np.bool_,
    )
    expected = np.expand_dims(expected, axis=1)

    gamma = 4
    assert np.array_equal(expected, hshap_features(gamma))


def test_w():
    gamma = 4

    c = 0
    expected = 6 / 24
    assert w(c, gamma) == approx(expected)

    c = 1
    expected = 2 / 24
    assert w(c, gamma) == approx(expected)

    c = 2
    expected = 2 / 24
    assert w(c, gamma) == approx(expected)

    c = 3
    expected = 6 / 24
    assert w(c, gamma) == approx(expected)


def test_shapley_matrix():
    gamma = 3
    device = (
        torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
    )

    with raises(NotImplementedError, match="Only implemented for gamma = 4"):
        shapley_matrix(gamma, device)

    gamma = 4
    assert shapley_matrix(gamma, device).device == device

    assert shapley_matrix(gamma, device).size() == (2 ** gamma, gamma)

    expected_first_column = torch.tensor(
        [
            -6 / 24,
            6 / 24,
            -2 / 24,
            -2 / 24,
            -2 / 24,
            2 / 24,
            2 / 24,
            2 / 24,
            -2 / 24,
            -2 / 24,
            -2 / 24,
            2 / 24,
            2 / 24,
            2 / 24,
            -6 / 24,
            6 / 24,
        ],
        device=device,
    )
    assert torch.eq(expected_first_column, shapley_matrix(gamma, device)[:, 0]).all()


def test_mask_features_():
    c, h, w = 3, 4, 8
    mask = torch.zeros(5, c, h, w, dtype=torch.bool)
    root_coords = np.array([[0, 0], [h, w]], dtype=np.uint16)

    expected = mask.clone()
    expected[1, :, 0:2, 0:4].fill_(True)
    expected[2, :, 0:2, 4:8].fill_(True)
    expected[3, :, 2:4, 0:4].fill_(True)
    expected[4, :, 2:4, 4:8].fill_(True)

    mask_features_(mask, root_coords)

    assert torch.eq(expected, mask).all()


def test_mask_input_():
    c, h, w = 3, 4, 8
    input = torch.ones(c, h, w)
    background = torch.zeros(c, h, w)
    root_coords = np.array([[0, 0], [h, w]], dtype=np.uint16)

    # first feature
    path_0 = np.array([1, 0, 0, 0])
    expected_0 = background.clone()
    expected_0[:, 0:2, 0:4].fill_(1)

    input_0 = input.clone()
    root_coords_0 = root_coords.copy()
    mask_input_(input_0, path_0, background, root_coords_0)
    assert torch.eq(expected_0, input_0).all()
    assert np.array_equal(root_coords_0, np.array([[0, 0], [2, 4]]))

    # second feature
    path_1 = np.array([0, 1, 0, 0])
    expected_1 = background.clone()
    expected_1[:, 0:2, 4:8].fill_(1)

    input_1 = input.clone()
    root_coords_1 = root_coords.copy()
    mask_input_(input_1, path_1, background, root_coords_1)
    assert torch.eq(expected_1, input_1).all()
    assert np.array_equal(root_coords_1, np.array([[0, 4], [2, 8]]))

    # third feature
    path_2 = np.array([0, 0, 1, 0])
    expected_2 = background.clone()
    expected_2[:, 2:4, 0:4].fill_(1)

    input_2 = input.clone()
    root_coords_2 = root_coords.copy()
    mask_input_(input_2, path_2, background, root_coords_2)
    assert torch.eq(expected_2, input_2).all()
    assert np.array_equal(root_coords_2, np.array([[2, 0], [4, 4]]))

    # fourth feature
    path_3 = np.array([0, 0, 0, 1])
    expected_3 = background.clone()
    expected_3[:, 2:4, 4:8].fill_(1)

    input_3 = input.clone()
    root_coords_3 = root_coords.copy()
    mask_input_(input_3, path_3, background, root_coords_3)
    assert torch.eq(expected_3, input_3).all()
    assert np.array_equal(root_coords_3, np.array([[2, 4], [4, 8]]))


def test_mask_map():
    c, h, w = 3, 4, 8
    map = torch.zeros(c, h, w)
    score = 1
    root_coords = np.array([[0, 0], [h, w]], dtype=np.uint16)

    # first feature
    path_0 = np.array([1, 0, 0, 0])
    expected_0 = map.clone()
    expected_0[:, 0:2, 0:4].fill_(score)
    mask_map_(map, path_0, score, root_coords)

    assert torch.eq(expected_0, map).all()

    # second feature
    path_1 = np.array([0, 1, 0, 0])
    expected_1 = map.clone()
    expected_1[:, 0:2, 4:8].fill_(score)
    mask_map_(map, path_1, score, root_coords)

    assert torch.eq(expected_1, map).all()

    # third feature
    path_2 = np.array([0, 0, 1, 0])
    expected_2 = map.clone()
    expected_2[:, 2:4, 0:4].fill_(score)
    mask_map_(map, path_2, score, root_coords)

    assert torch.eq(expected_2, map).all()

    # fourth feature
    path_3 = np.array([0, 0, 0, 1])
    expected_3 = map.clone()
    expected_3[:, 2:4, 4:8].fill_(score)
    mask_map_(map, path_3, score, root_coords)

    assert torch.eq(expected_3, map).all()



import torch
import numpy as np
from torch import Tensor

factorial = np.math.factorial


def hshap_features(gamma: int) -> np.ndarray:
    return np.expand_dims(np.eye(gamma, dtype=np.bool_), axis=1)


def w(c: int, gamma: int) -> int:
    return factorial(c) * factorial(gamma - c - 1) / factorial(gamma)


def shapley_matrix(gamma: int, device: torch.device) -> Tensor:
    if gamma != 4:
        raise NotImplementedError("Only implemented for gamma = 4")
    # construct matrix as copies of first row
    W = torch.tensor(
        [
            [-w(0, gamma), w(0, gamma)]
            + 3 * [-w(1, gamma)]
            + 3 * [w(1, gamma)]
            + 3 * [-w(2, gamma)]
            + 3 * [w(2, gamma)]
            + [-w(3, gamma), w(3, gamma)]
        ],
        device=device,
    ).repeat(gamma, 1)
    # update second row
    W[1, 1] = -w(1, gamma)
    W[1, 2] = w(0, gamma)
    W[1, 6] = -w(2, gamma)
    W[1, 7] = -w(2, gamma)
    W[1, 8] = w(1, gamma)
    W[1, 9] = w(1, gamma)
    W[1, -3] = -w(3, gamma)
    W[1, -2] = w(2, gamma)
    # update second row
    W[2, 1] = -w(1, gamma)
    W[2, 3] = w(0, gamma)
    W[2, 5] = -w(2, gamma)
    W[2, 7] = -w(2, gamma)
    W[2, 8] = w(1, gamma)
    W[2, 10] = w(1, gamma)
    W[2, -4] = -w(3, gamma)
    W[2, -2] = w(2, gamma)
    # update third row
    W[3, 1] = -w(1, gamma)
    W[3, 4] = w(0, gamma)
    W[3, 5] = -w(2, gamma)
    W[3, 6] = -w(2, gamma)
    W[3, 9] = w(1, gamma)
    W[3, 10] = w(1, gamma)
    W[3, -5] = -w(3, gamma)
    W[3, -2] = w(2, gamma)

    return W.transpose(0, 1)


def mask_features_(
    feature_mask: Tensor,
    root_coords: np.ndarray,
) -> None:
    center = np.mean(root_coords, axis=0, dtype=np.uint16)

    feature_mask[
        1, :, root_coords[0, 0] : center[0], root_coords[0, 1] : center[1]
    ].fill_(True)
    feature_mask[
        2, :, root_coords[0, 0] : center[0], center[1] : root_coords[1, 1]
    ].fill_(True)
    feature_mask[
        3, :, center[0] : root_coords[1, 0], root_coords[0, 1] : center[1]
    ].fill_(True)
    feature_mask[
        4, :, center[0] : root_coords[1, 0], center[1] : root_coords[1, 1]
    ].fill_(True)


def mask_input_(
    input: Tensor,
    path: np.ndarray,
    background: Tensor,
    root_coords: np.ndarray,
) -> None:
    if not np.all(path):
        center = np.mean(root_coords, axis=0, dtype=np.uint16)

        if not path[0]:
            input[
                :, root_coords[0, 0] : center[0], root_coords[0, 1] : center[1]
            ] = background[
                :, root_coords[0, 0] : center[0], root_coords[0, 1] : center[1]
            ]
        if not path[1]:
            input[
                :, root_coords[0, 0] : center[0], center[1] : root_coords[1, 1]
            ] = background[
                :, root_coords[0, 0] : center[0], center[1] : root_coords[1, 1]
            ]
        if not path[2]:
            input[
                :, center[0] : root_coords[1, 0], root_coords[0, 1] : center[1]
            ] = background[
                :, center[0] : root_coords[1, 0], root_coords[0, 1] : center[1]
            ]
        if not path[3]:
            input[
                :, center[0] : root_coords[1, 0], center[1] : root_coords[1, 1]
            ] = background[
                :, center[0] : root_coords[1, 0], center[1] : root_coords[1, 1]
            ]

        feature_id = np.nonzero(path)[0][0]
        feature_row, feature_column = feature_id // 2, feature_id % 2
        root_coords[0, 0] = center[0] if feature_row == 1 else root_coords[0, 0]
        root_coords[0, 1] = center[1] if feature_column == 1 else root_coords[0, 1]
        root_coords[1, 0] = center[0] if (1 - feature_row) == 1 else root_coords[1, 0]
        root_coords[1, 1] = (
            center[1] if (1 - feature_column) == 1 else root_coords[1, 1]
        )


def mask_map_(
    map: Tensor,
    path: np.ndarray,
    score: float,
    root_coords: np.ndarray,
):
    center = np.mean(root_coords, axis=0, dtype=np.uint16)

    if path[0]:
        map[:, root_coords[0, 0] : center[0], root_coords[0, 1] : center[1]].fill_(
            score
        )
    elif path[1]:
        map[:, root_coords[0, 0] : center[0], center[1] : root_coords[1, 1]].fill_(
            score
        )
    elif path[2]:
        map[:, center[0] : root_coords[1, 0], root_coords[0, 1] : center[1]].fill_(
            score
        )
    elif path[3]:
        map[:, center[0] : root_coords[1, 0], center[1] : root_coords[1, 1]].fill_(
            score
        )



from setuptools import setup, find_packages

VERSION = "0.1.4"
PACKAGE_NAME = "h-shap"
DESCRIPTION = "Fast Hierarchical Games for Image Explanations"
LONG_DESCRIPTION = "`h-shap` provides a fast, hierarchical implementation of Shapley coefficients for image explanations. It is exact, and it does not rely on approximation. In binary classification scenarios, `h-shap` guarantees an exponential computational advantage when explaining an important concept contained in the image."
AUTHOR = "Jacopo Teneggi"
AUTHOR_EMAIL = "jtenegg1@jhu.edu"

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="https://github.com/Sulam-Group/h-shap",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license="MIT",
    packages=find_packages(),
    install_requires=["numpy", "torch"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)
